"""
Django settings for arkindex project.

Generated by 'django-admin startproject' using Django 1.11.6.

For more information on this file, see
https://docs.djangoproject.com/en/1.11/topics/settings/

For the full list of settings and their values, see
https://docs.djangoproject.com/en/1.11/ref/settings/
"""

import os
import sys
import warnings
from datetime import timedelta
from pathlib import Path

from corsheaders.defaults import default_headers

from arkindex.project.config import CacheType, get_settings_parser

# Build paths inside the project like this: BASE_DIR / ...
_base_dir = os.environ.get("BASE_DIR")
BASE_DIR = Path(_base_dir) if _base_dir else Path(__file__).resolve().parent.parent

# Used for special cases during configuration parsing and settings loading
TEST_ENV = "test" in sys.argv

# Matches arkindex shell[_plus] to disable Sentry reporting
DJANGO_SHELL = len(sys.argv) > 1 and sys.argv[1] in ("shell", "shell_plus")

CONFIG_PATH = Path(os.environ.get("CONFIG_PATH", BASE_DIR / "config.yml"))
parser = get_settings_parser(BASE_DIR)
conf = parser.parse(CONFIG_PATH, exist_ok=True)

# Read Version either from Docker static file or local file
_version = "/etc/arkindex.version" \
    if os.path.exists("/etc/arkindex.version") \
    else BASE_DIR.parent / "VERSION"
with open(_version) as f:
    VERSION = f.read().strip()

ARKINDEX_ENV = conf["arkindex_env"]
SECRET_KEY = conf["secret_key"]
LOCAL_IMAGESERVER_ID = conf["local_imageserver_id"]

ALLOWED_HOSTS = conf["allowed_hosts"]
# Docker container name resolution
ALLOWED_HOSTS += ["127.0.0.1", "localhost", "backend", "ark-backend"]

# How long, in seconds, before a WorkerActivity in a `started` state can be retried
WORKER_ACTIVITY_TIMEOUT = conf["worker_activity_timeout"]

PUBLIC_HOSTNAME = conf["public_hostname"]

PROMETHEUS_METRICS_PORT = conf["metrics_port"]


# Database
def _conf_to_django_db(config):
    """Turn a YAML database config into a Django database config"""
    return {
        "ENGINE": "django.contrib.gis.db.backends.postgis",
        "NAME": config["name"],
        "USER": config["user"],
        "PASSWORD": config["password"],
        "HOST": config["host"],
        "PORT": config["port"],
    }


_primary_db = _conf_to_django_db(conf["database"])
DATABASES = {"default": _primary_db}

if conf["database"]["replica"] is not None:
    # Make the replica's options default to the primary's
    _replica_db = _conf_to_django_db(conf["database"]["replica"])
    for key, value in _primary_db.items():
        if not _replica_db.get(key):
            _replica_db[key] = value

    DATABASES["replica"] = _replica_db
    DATABASE_ROUTERS = ["arkindex.project.routers.PrimaryReplicaRouter"]

DEFAULT_AUTO_FIELD = "django.db.models.AutoField"

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = ARKINDEX_ENV == "dev"

# Run all system checks when starting the server
ALL_CHECKS = os.environ.get("ALL_CHECKS") == "true" or "runserver" in sys.argv

# Required for django-debug-toolbar
INTERNAL_IPS = ["127.0.0.1", "127.0.1.1"]

# Application definition
INSTALLED_APPS = [
    "django.contrib.admin",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.messages",
    "django.contrib.staticfiles",
    "django.contrib.postgres",
    "django_admin_hstore_widget",

    # Tools
    "rest_framework",
    "rest_framework.authtoken",
    "corsheaders",
    "django_rq",
    "drf_spectacular",
    "pgtrigger",

    # Our apps
    "arkindex.ponos",
    "arkindex.images",
    "arkindex.documents",
    "arkindex.users",
    "arkindex.process",
    "arkindex.training",
]

MIDDLEWARE = [
    "corsheaders.middleware.CorsMiddleware",
    "django.middleware.security.SecurityMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
]

ROOT_URLCONF = "arkindex.project.urls"

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [
            os.path.join(BASE_DIR, "templates"),
        ],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
            ],
        },
    },
]

WSGI_APPLICATION = "arkindex.project.wsgi.application"
ASGI_APPLICATION = "arkindex.project.routing.application"


# Password validation
# https://docs.djangoproject.com/en/1.11/ref/settings/#auth-password-validators

AUTH_PASSWORD_VALIDATORS = [
    {
        "NAME": "django.contrib.auth.password_validation.UserAttributeSimilarityValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.MinimumLengthValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.CommonPasswordValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.NumericPasswordValidator",
    },
]

AUTH_USER_MODEL = "users.User"

# Internationalization
# https://docs.djangoproject.com/en/1.11/topics/i18n/

LANGUAGE_CODE = "en-us"

TIME_ZONE = "UTC"

USE_I18N = True

USE_TZ = True


# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/1.11/howto/static-files/
STATIC_URL = "/static/"
STATIC_ROOT = conf["static"]["root_path"]
STATICFILES_DIRS = []

# By default the frontend version is the same as the backend
FRONTEND_VERSION = conf["static"].get("frontend_version") or VERSION

# API
REST_FRAMEWORK = {
    "DEFAULT_AUTHENTICATION_CLASSES": (
        "rest_framework.authentication.SessionAuthentication",
        "rest_framework.authentication.TokenAuthentication",
        "arkindex.ponos.authentication.TaskAuthentication",
    ),
    "DEFAULT_PAGINATION_CLASS": "arkindex.project.pagination.PageNumberPagination",
    "DEFAULT_SCHEMA_CLASS": "arkindex.project.openapi.AutoSchema",
    "PAGE_SIZE": 20,
}

SIMPLE_JWT = {
    "USER_ID_CLAIM": "agent_id",
    "ROTATE_REFRESH_TOKENS": True,
    "ACCESS_TOKEN_LIFETIME": timedelta(hours=6),
    "SIGNING_KEY": conf["jwt_signing_key"] or SECRET_KEY,
}

SPECTACULAR_SETTINGS = {
    "CAMELIZE_NAMES": True,
    "SCHEMA_PATH_PREFIX": "/api/v1/",
    "TITLE": "Arkindex API",
    "CONTACT": {
        "name": "Teklia",
        "url": "https://www.teklia.com/",
        "email": "contact@teklia.com",
    },
    "VERSION": VERSION,
    "EXTERNAL_DOCS": {
        "url": "https://doc.arkindex.org/",
        "description": "Arkindex documentation"
    },
    "POSTPROCESSING_HOOKS": [
        "drf_spectacular.hooks.postprocess_schema_enums",
        "arkindex.project.openapi.operation_id_pascal_case",
        "arkindex.project.openapi.check_tags",
        "arkindex.project.openapi.request_body_x_name",
    ],
    # Spectacular generates enum names from field names, so if two different enums are linked
    # to the same field name (e.g. MetaType and EntityType on `type` fields of different serializers),
    # Spectacular can warn about duplicates.
    # To fix this, the entries below map custom names to each affected enum.
    "ENUM_NAME_OVERRIDES": {
        "MetaType": "arkindex.documents.models.MetaType",
        "ProcessMode": "arkindex.process.models.ProcessMode",
        "S3FileStatus": "arkindex.project.aws.S3FileStatus",
        "ClassificationState": "arkindex.documents.models.ClassificationState",
        "PonosState": "arkindex.ponos.models.State",
        "WorkerVersionState": "arkindex.process.models.WorkerVersionState",
        "ModelVersionState": "arkindex.training.models.ModelVersionState",
        "MetricMode": "arkindex.training.models.MetricMode",
        "DatasetState": "arkindex.training.models.DatasetState",
        "FeatureUsage": "arkindex.process.models.FeatureUsage",
    },
    "TAGS": [
        {"name": "classifications"},
        {"name": "corpora"},
        {"name": "elements"},
        {"name": "entities"},
        {"name": "exports"},
        {"name": "files"},
        {
            "name": "iiif",
            "description": "IIIF manifests, annotation lists and services",
        },
        {"name": "process"},
        {"name": "ingest"},
        {"name": "images"},
        {"name": "jobs"},
        {
            "name": "ml",
            "description": "Machine Learning tools and results",
        },
        {
            "name": "training",
            "description": "Machine Learning models training",
        },
        {"name": "datasets"},
        {"name": "ponos"},
        {"name": "repos"},
        {"name": "search"},
        {"name": "transcriptions"},
        {"name": "users"},
    ]
}

# Solr config
SOLR_API_URL = conf["solr"]["api_url"]
SOLR_PAGINATION_SIZE = 20

SEARCH_FILTER_MAX_TERMS = 10

# Use SSL proxy
SECURE_PROXY_SSL_HEADER = ("HTTP_X_FORWARDED_PROTO", "https")

# IIIF manifests
IIIF_PRESENTATION_CONTEXT = "http://iiif.io/api/presentation/2/context.json"
IIIF_IMAGE_CONTEXT = "http://iiif.io/api/image/2/context.json"
IIIF_IMAGE_SERVICE_PROFILE = "http://iiif.io/api/image/2/level2.json"

# IIIF manifest download timeout
# See http://docs.python-requests.org/en/master/user/advanced/#timeouts
IIIF_DOWNLOAD_TIMEOUT = (30, 60)

# check_images sample size when checking all servers
CHECK_IMAGES_SAMPLE_SIZE = 20


if conf["cache"]["type"] is None:
    conf["cache"]["type"] = CacheType.Dummy if DEBUG else CacheType.Memory

if conf["cache"]["type"] == CacheType.Redis:
    CACHES = {
        "default": {
            "BACKEND": "django.core.cache.backends.redis.RedisCache",
            "LOCATION": conf["cache"]["url"],
        }
    }
elif conf["cache"]["type"] == CacheType.Filesystem:
    CACHES = {
        "default": {
            "BACKEND": "django.core.cache.backends.filebased.FileBasedCache",
            "LOCATION": str(conf["cache"]["path"]),
        }
    }
elif conf["cache"]["type"] == CacheType.Memory:
    CACHES = {
        "default": {
            "BACKEND": "django.core.cache.backends.locmem.LocMemCache"
        }
    }
elif conf["cache"]["type"] == CacheType.Dummy:
    CACHES = {
        "default": {
            "BACKEND": "django.core.cache.backends.dummy.DummyCache"
        }
    }

_rq_queue_conf = {
    "HOST": conf["redis"]["host"],
    "PORT": conf["redis"]["port"],
    "DB": conf["redis"]["db"],
    "PASSWORD": conf["redis"]["password"],
    "DEFAULT_TIMEOUT": conf["redis"]["timeout"],
}
RQ_QUEUES = {
    "default": _rq_queue_conf,
    "high": _rq_queue_conf,
    "tasks": _rq_queue_conf,
}

RQ_TIMEOUTS = conf["job_timeouts"]

RQ = {
    "JOB_CLASS": "arkindex.project.rq_overrides.Job",
    "QUEUE_CLASS": "arkindex.project.rq_overrides.Queue"
}

# How many keys to delete at once inside a sorted set in Redis using a single ZREM command
REDIS_ZREM_CHUNK_SIZE = 10000

# How long before a corpus export can be run again after a successful one
EXPORT_TTL_SECONDS = conf["export"]["ttl"]
# Available database sources for corpus exports
EXPORT_SOURCES = ["default"]

LOGGING = {
    "version": 1,
    "filters": {
        "require_debug_false": {
            "()": "django.utils.log.RequireDebugFalse",
        },
        "require_debug_true": {
            "()": "django.utils.log.RequireDebugTrue",
        },
    },
    "handlers": {
        "console": {
            "level": "DEBUG",
            "class": "logging.StreamHandler",
            "formatter": "simple"
        },
        "console_debug": {
            "level": "DEBUG",
            "class": "logging.StreamHandler",
            "formatter": "simple",
            "filters": ["require_debug_true"],
        },
    },
    "loggers": {
        "django": {
            "handlers": ["console"],
            "level": "INFO",
            "propagate": True,
        },
        "arkindex": {
            "handlers": ["console"],
            "level": "INFO",
            "propagate": True,
        },
        "rq": {
            "handlers": ["console"],
            "level": "INFO",
            "propagate": True,
        }
    },
    "formatters": {
        "simple": {
            "format": "%(asctime)s [%(levelname)s] %(message)s"
        },
    },
}

# Email
EMAIL_SUBJECT_PREFIX = "[Arkindex {}] ".format(ARKINDEX_ENV)
if conf["email"]:
    ADMINS = [("", address) for address in conf["email"]["error_report_recipients"]]
    EMAIL_BACKEND = "django.core.mail.backends.smtp.EmailBackend"
    EMAIL_HOST = conf["email"]["host"]
    EMAIL_PORT = conf["email"]["port"]
    EMAIL_HOST_USER = conf["email"]["user"]
    DEFAULT_FROM_EMAIL = SERVER_EMAIL = EMAIL_HOST_USER
    EMAIL_HOST_PASSWORD = conf["email"]["password"]
    EMAIL_USE_TLS = True
else:
    EMAIL_BACKEND = "django.core.mail.backends.console.EmailBackend"


# Cookies
CSRF_COOKIE_NAME = conf["csrf"]["cookie_name"]
CSRF_COOKIE_DOMAIN = conf["csrf"]["cookie_domain"]
CSRF_COOKIE_SAMESITE = conf["csrf"]["cookie_samesite"].value
CSRF_COOKIE_SECURE = conf["csrf"]["cookie_secure"]
CSRF_TRUSTED_ORIGINS = conf["csrf"]["trusted_origins"]

SESSION_COOKIE_NAME = conf["session"]["cookie_name"]
SESSION_COOKIE_DOMAIN = conf["session"]["cookie_domain"]
SESSION_COOKIE_SAMESITE = conf["session"]["cookie_samesite"].value
SESSION_COOKIE_SECURE = conf["session"]["cookie_secure"]

# When a real cache system is available, write the session to the cache
# and to the database (write-through), so that the backend can immediately checks
# in the cache, and then fallback to the DB. It's required when using a Postgresql
# cluster, or a multi backend setup
if conf["cache"]["type"] == CacheType.Redis:
    SESSION_ENGINE = "django.contrib.sessions.backends.cached_db"
else:
    SESSION_ENGINE = "django.contrib.sessions.backends.db"

CORS_ALLOWED_ORIGINS = conf["cors"]["origin_whitelist"]
CORS_ALLOW_CREDENTIALS = True
CORS_ALLOW_HEADERS = default_headers + (
    "cache-control",  # Allow the frontend to prevent the browser from caching some API requests
    "if-modified-since",  # Allow frontend cache invalidation
)
CORS_URLS_REGEX = r"^/(api|ponos)/.*$"

# Support CORS suffixes
if conf["cors"]["suffixes"]:
    CORS_ALLOWED_ORIGIN_REGEXES = [
        r"^https://.+{}".format(suffix)
        for suffix in conf["cors"]["suffixes"]
    ]

# Amazon S3
AWS_ACCESS_KEY = conf["s3"]["access_key_id"]
AWS_SECRET_KEY = conf["s3"]["secret_access_key"]
AWS_ENDPOINT = conf["s3"]["endpoint"]
AWS_REGION = conf["s3"]["region"]
PONOS_S3_LOGS_BUCKET = conf["s3"]["ponos_logs_bucket"]
PONOS_S3_ARTIFACTS_BUCKET = conf["s3"]["ponos_artifacts_bucket"]

AWS_THUMBNAIL_BUCKET = conf["s3"]["thumbnails_bucket"]
AWS_STAGING_BUCKET = conf["s3"]["staging_bucket"]
AWS_EXPORT_BUCKET = conf["s3"]["export_bucket"]
AWS_TRAINING_BUCKET = conf["s3"]["training_bucket"]

INGEST_S3_ACCESS_KEY = conf["ingest"]["access_key_id"]
INGEST_S3_SECRET_KEY = conf["ingest"]["secret_access_key"]
INGEST_S3_ENDPOINT = conf["ingest"]["endpoint"]
INGEST_S3_REGION = conf["ingest"]["region"]
INGEST_IMAGESERVER_ID = conf["ingest"]["imageserver_id"]
INGEST_EXTRA_BUCKETS = conf["ingest"]["extra_buckets"]
INGEST_PREFIX_BY_BUCKET_NAME = conf["ingest"]["prefix_by_bucket_name"]

# Ponos integration
_ponos_env = {
    "ARKINDEX_API_CSRF_COOKIE": CSRF_COOKIE_NAME
}
if DEBUG:
    # In dev, include overridable API info
    _ponos_env.update({
        "ARKINDEX_API_URL": "http://localhost:8000/api/v1/",
        "ARKINDEX_API_TOKEN": "deadbeefTestToken",
    })
_ponos_env.update(conf["ponos"]["default_env"])
PONOS_DEFAULT_ENV = _ponos_env
PONOS_ARTIFACT_MAX_SIZE = conf["ponos"]["artifact_max_size"]
PONOS_RQ_EXECUTION = True
# Base data directory for RQ tasks execution (in the docker container)
PONOS_DATA_DIR = "/data"

# Docker images used by our ponos workflow
ARKINDEX_TASKS_IMAGE = conf["docker"]["tasks_image"]

# Robots.txt options
ROBOTS_TXT_DISALLOW = conf["robots_txt_disallow"]

# Processes options
IMPORTS_WORKER_VERSION = conf["imports_worker_version"]
MAX_CHUNKS = conf["workers_max_chunks"]

# Archive formats supported in file imports
ARCHIVE_MIME_TYPES = {
    # .zip, standard
    "application/zip",
    # .zip, from Windows only
    "application/x-zip-compressed",
    # .tar
    "application/x-tar",
    # .tar.gz, from libmagic only
    "application/x-gtar",
    # .tgz, from browsers only
    "application/x-compressed-tar",
    # .gz from anywhere, .tgz from libmagic only
    "application/gzip",
    # .bz2, from browsers only
    "application/x-bzip",
    # .bz2, from libmagic only
    "application/x-bzip2",
    # .lzma
    "application/x-lzma",
    # .xz
    "application/x-xz",
    # .zst
    "application/zstd",
}

# User groups with special permissions
# Deprecated, left there only to run the users.0004 migration
INTERNAL_GROUP_ID = conf["internal_group_id"]

# CDN Assets URL to use for arkindex remote CSS/JS/Images assets
CDN_ASSETS_URL = conf["static"]["cdn_assets_url"]
if CDN_ASSETS_URL is not None:
    CDN_ASSETS_URL = CDN_ASSETS_URL.rstrip("/")
    STATIC_URL = f"{CDN_ASSETS_URL}/{VERSION}/static/"

MIRADOR_URL = conf["static"]["mirador_url"]
UNIVERSAL_VIEWER_URL = conf["static"]["universal_viewer_url"]

# Feature flags
ARKINDEX_FEATURES = {
    **conf["features"],
    # Enable doorbell feature if credentials are defined
    "doorbell": bool(conf["doorbell"].get("id") and conf["doorbell"].get("appkey"))
}

SENTRY_DSN = conf["sentry"]["dsn"]
FRONTEND_SENTRY_DSN = conf["sentry"]["frontend_dsn"]

# https://doorbell.io settings
FRONTEND_DOORBELL = conf["doorbell"]

# Banner settings
FRONTEND_BANNER = conf["banner"]

WORKER_CLEANUP_DELAY = conf["cleanup"]["worker_delay"]
MODEL_CLEANUP_DELAY = conf["cleanup"]["model_delay"]

# Optional default group to add new users to when they complete email verification
SIGNUP_DEFAULT_GROUP = conf["signup_default_group"]

SQL_VALIDATION_DIR = BASE_DIR / "sql_validation"

if TEST_ENV:
    # Overrides for unit tests
    AWS_ACCESS_KEY = "test"
    AWS_SECRET_KEY = "test"
    AWS_ENDPOINT = "http://s3"
    PONOS_S3_ARTIFACTS_BUCKET = "ponos-artifacts"
    PONOS_S3_LOGS_BUCKET = "ponos-logs"
    LOCAL_IMAGESERVER_ID = 1
    SIGNUP_DEFAULT_GROUP = None

    # Causes RQ tasks to run on the main thread
    for queue in RQ_QUEUES.values():
        queue["ASYNC"] = False

    # Turn Django's UnorderedObjectListWarning into exceptions
    warnings.filterwarnings("error", category=RuntimeWarning, module="django.core.paginator")
    warnings.filterwarnings("error", category=RuntimeWarning, module="rest_framework.pagination")

    # Disable RQ tasks scheduler during tests
    PONOS_RQ_EXECUTION = False

# Optional unit tests runner with code coverage
try:
    import xmlrunner  # noqa
    TEST_RUNNER = "xmlrunner.extra.djangotestrunner.XMLTestRunner"
    TEST_OUTPUT_FILE_NAME = "test-report.xml"
except ImportError:
    pass

# Local settings
if DEBUG and not TEST_ENV:
    try:
        from .local_settings import *  # noqa
    except ImportError:
        pass

try:
    import django_extensions  # noqa
    INSTALLED_APPS.append("django_extensions")
    # With disable_existing_loggers=True, Django's logging config causes arkindex runserver_plus --print-sql
    # to no longer be able to log SQL queries.
    # https://github.com/django-extensions/django-extensions/issues/1626#issuecomment-774698668
    LOGGING["loggers"]["django_extensions.management.commands.runserver_plus"] = {
        "handlers": ["console"],
        "level": "INFO",
    }
except ImportError:
    pass

try:
    import debug_toolbar.settings  # noqa
    MIDDLEWARE.insert(0, "debug_toolbar.middleware.DebugToolbarMiddleware")
    INSTALLED_APPS.append("debug_toolbar")
except ImportError:
    pass

if SENTRY_DSN and not DJANGO_SHELL:
    import sentry_sdk
    from sentry_sdk.integrations.django import DjangoIntegration
    from sentry_sdk.integrations.logging import ignore_logger
    from sentry_sdk.integrations.redis import RedisIntegration
    from sentry_sdk.integrations.rq import RqIntegration
    ignore_logger("django.security.DisallowedHost")

    sentry_sdk.init(
        dsn=SENTRY_DSN,
        environment=ARKINDEX_ENV,
        release=f"arkindex-backend@{VERSION}",
        debug=DEBUG,
        send_default_pii=True,
        integrations=[
            DjangoIntegration(),
            RedisIntegration(),
            RqIntegration(),
        ],
    )
